###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2026 - Yan Georget
###############################################################################
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import EVENT_MASK_MIN_MAX, MAX, MIN, PROP_CONSISTENCY, PROP_INCONSISTENCY


def get_complexity_affine_eq(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return n


@njit(cache=True)
def get_triggers_affine_eq(n: int, variable: int, parameters: NDArray) -> int:
    """
    This propagator is triggered whenever there is a change in the domain of a variable.
    :param parameters: the parameters, unused here
    :return: an array of triggers
    """
    return EVENT_MASK_MIN_MAX


@njit(cache=True)
def compute_domains_affine_eq(domains: NDArray, parameters: NDArray) -> int:
    """
    Implements Sigma_i a_i * x_i = a_{n}.
    :param domains: the domains of the variables, x is an alias for domains
    :param parameters: the parameters of the propagator, a is an alias for parameters
    :return: the status of the propagation (consistency, inconsistency or entailment) as an int
    """
    factors = parameters[:-1]
    n = len(factors)
    has_changed = True
    while has_changed:
        has_changed = False
        domain_sum_min = domain_sum_max = -parameters[-1]
        for i in range(n):
            factor = factors[i]
            if factor > 0:
                domain_sum_min += factor * domains[i, MAX]
                domain_sum_max += factor * domains[i, MIN]
            elif factor < 0:
                domain_sum_min += factor * domains[i, MIN]
                domain_sum_max += factor * domains[i, MAX]
        for i in range(n):
            factor = factors[i]
            if factor != 0:
                if factor > 0:
                    new_min = domains[i, MAX] - (-domain_sum_min // -factor)
                    new_max = domains[i, MIN] + (-domain_sum_max // factor)
                else:
                    new_min = domains[i, MAX] - (domain_sum_max // factor)
                    new_max = domains[i, MIN] + (domain_sum_min // -factor)
                if new_min > domains[i, MIN]:
                    domains[i, MIN] = new_min
                    has_changed = True
                if new_max < domains[i, MAX]:
                    domains[i, MAX] = new_max
                    has_changed = True
                if domains[i, MIN] > domains[i, MAX]:
                    return PROP_INCONSISTENCY
    return PROP_CONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2026 - Yan Georget
###############################################################################
import copy
import logging
from typing import Any, Iterable, List, Optional, Self, Sequence, Tuple, Union

import numpy as np
from numba import njit
from numpy.typing import NDArray
from rich import print

from nucs.constants import (
    EVENT_MASK_NB,
    NUMBA_DISABLE_JIT,
    PARAM,
    RANGE_END,
    RANGE_START,
    SIGNATURE_GET_TRIGGERS,
    TYPE_GET_TRIGGERS,
    VARIABLE,
)
from nucs.numba_helper import build_function_address_list, function_from_address
from nucs.propagators.propagators import GET_COMPLEXITY_FCTS, GET_TRIGGERS_FCTS

logger = logging.getLogger(__name__)


class Problem:
    """
    A problem is defined by:
    - a list of domains,
    - a list of propagators.
    A variable is a domain index.
    """

    def __init__(self, domains: Union[Iterable[Tuple[int, int]]]):
        """
        Initializes the problem.
        :param domains: the domains
        """
        self.domains = [(domain, domain) if isinstance(domain, int) else domain for domain in domains]
        self.domain_nb = len(self.domains)
        self.propagators: List[Tuple[List[int], int, List[int]]] = []
        self.propagator_nb = 0

    def split(self, split_nb: int, var: int) -> List[Self]:
        """
        Splits a problem into several sub-problems by splitting the domain of a variable.
        :param split_nb: the number of sub-problems
        :param var: the index of the variable
        :return: a list of sub-problems
        """
        logger.debug(f"Splitting in {split_nb} problems with variable {var}")
        domain = self.domains[var]
        domain_min, domain_max = domain
        domain_size = domain_max - domain_min + 1
        problems = []
        min_idx = domain_min
        for split_idx in range(split_nb):
            problem = copy.deepcopy(self)
            max_idx = min_idx + domain_size // split_nb - (0 if split_idx < domain_size % split_nb else 1)
            problem.domains[var] = (min_idx, max_idx)
            min_idx = max_idx + 1
            problems.append(problem)
        return problems

    def add_variable(self, domain: Union[int, Tuple[int, int]]) -> int:
        """
        Adds an extra variable to the problem.
        :param domain: the domain of the variable
        :return: the extra variable
        """
        var = len(self.domains)
        self.domains.append((domain, domain) if isinstance(domain, int) else domain)
        self.domain_nb = var + 1
        return var

    def add_variables(self, domains: Sequence[Union[int, Tuple[int, int]]]) -> int:
        """
        Adds extra variables to the problem.
        :param domains: the domains of the variables
        :return: the first added variable
        """
        var = len(self.domains)
        self.domains.extend([(domain, domain) if isinstance(domain, int) else domain for domain in domains])
        self.domain_nb = len(self.domains)
        return var

    def add_propagator(
        self, algorithm: int, variables: Iterable[int], parameters: Optional[Iterable[int]] = None
    ) -> None:
        """
        Adds an extra propagator.
        """
        parameters = [] if parameters is None else list(parameters)
        variables = list(variables)
        self.propagators.append((variables, algorithm, parameters))
        self.propagator_nb += 1

    def init(self) -> None:
        """
        Completes the initialization of the problem.
        """
        logger.debug("Initializing problem")
        # Sort the propagators based on their estimated amortized complexities.
        self.propagators.sort(
            key=lambda propagator: (
                GET_COMPLEXITY_FCTS[propagator[1]](len(propagator[0]), propagator[2]),
                propagator[1],
            )
        )
        self.algorithms = np.array([propagator[1] for propagator in self.propagators], dtype=np.uint8)
        # We will store propagator specific data in a global arrays, we need to compute variables and parameter bounds.
        logger.debug("Initializing bounds")
        self.bounds = np.zeros((max(1, self.propagator_nb), 2, 2), dtype=np.uint32)  # some redundancy here
        init_bounds(self.bounds, self.propagators)
        logger.debug("Initializing props")
        self.props_variables = np.empty(self.bounds[-1, VARIABLE, RANGE_END], dtype=np.uint32)
        self.props_parameters = np.empty(self.bounds[-1, PARAM, RANGE_END], dtype=np.int32)
        init_props(self.props_variables, self.props_parameters, self.bounds, self.propagators)
        logger.debug("Initializing triggers")
        self.triggers = np.full((self.domain_nb, EVENT_MASK_NB, self.propagator_nb + 1), -1, dtype=np.int32)
        get_triggers_addrs = (
            np.empty(0)
            if NUMBA_DISABLE_JIT
            else np.array(build_function_address_list(GET_TRIGGERS_FCTS, SIGNATURE_GET_TRIGGERS))
        )
        init_triggers(
            self.triggers,
            self.domain_nb,
            self.propagator_nb,
            self.bounds,
            self.props_variables,
            self.props_parameters,
            self.algorithms,
            get_triggers_addrs,
        )
        logger.debug("Problem initialized")
        logger.info(f"Problem has {self.propagator_nb} propagators")
        logger.info(f"Problem has {self.domain_nb} variables")

    def solution_as_printable(self, solution: NDArray) -> Any:
        return solution.tolist()

    def print_solution(self, solution: Optional[NDArray]) -> None:
        if solution is not None:
            print(self.solution_as_printable(solution))
        else:
            print("No solution")


def init_bounds(bounds: NDArray, propagators: List[Tuple[List[int], int, List[int]]]) -> None:
    for prop_idx, prop in enumerate(propagators):
        if prop_idx > 0:
            bounds[prop_idx, :, RANGE_START] = bounds[prop_idx - 1, :, RANGE_END]
        bounds[prop_idx, VARIABLE, RANGE_END] = bounds[prop_idx, VARIABLE, RANGE_START] + len(prop[0])
        bounds[prop_idx, PARAM, RANGE_END] = bounds[prop_idx, PARAM, RANGE_START] + len(prop[2])


def init_props(
    props_variables: NDArray,
    props_parameters: NDArray,
    bounds: NDArray,
    propagators: List[Tuple[List[int], int, List[int]]],
) -> None:
    for prop_idx, prop in enumerate(propagators):
        props_variables[bounds[prop_idx, VARIABLE, RANGE_START] : bounds[prop_idx, VARIABLE, RANGE_END]] = prop[0]
        props_parameters[bounds[prop_idx, PARAM, RANGE_START] : bounds[prop_idx, PARAM, RANGE_END]] = prop[2]


@njit(cache=True)
def init_triggers(
    triggers: NDArray,
    domain_nb: int,
    propagator_nb: int,
    bounds: NDArray,
    props_variables: NDArray,
    props_parameters: NDArray,
    algorithms: NDArray,
    get_triggers_addrs: NDArray,
) -> None:
    # for each domain and event, we store the list of propagator indices followed by -1
    indices = np.zeros((domain_nb, EVENT_MASK_NB), dtype=np.uint32)
    if NUMBA_DISABLE_JIT:
        get_trigger_fcts = [GET_TRIGGERS_FCTS[algorithms[prop_idx]] for prop_idx in range(propagator_nb)]
    else:
        previous_algorithm = algorithms[0]
        previous_function = function_from_address(TYPE_GET_TRIGGERS, get_triggers_addrs[previous_algorithm])
        get_trigger_fcts = [previous_function] * propagator_nb  # necessary for numba to get the type
        for prop_idx in range(1, propagator_nb):
            if algorithms[prop_idx] == previous_algorithm:
                get_trigger_fcts[prop_idx] = previous_function
            else:
                previous_algorithm = algorithms[prop_idx]
                previous_function = get_trigger_fcts[prop_idx] = function_from_address(
                    TYPE_GET_TRIGGERS, get_triggers_addrs[previous_algorithm]
                )
    for prop_idx in range(propagator_nb):
        parameters = props_parameters[bounds[prop_idx, PARAM, RANGE_START] : bounds[prop_idx, PARAM, RANGE_END]]
        var_start = bounds[prop_idx, VARIABLE, RANGE_START]
        var_end = bounds[prop_idx, VARIABLE, RANGE_END]
        var_nb = var_end - var_start
        for var in range(var_start, var_end):
            trigger = get_trigger_fcts[prop_idx](var_nb, var - var_start, parameters)
            variable = props_variables[var]
            for event_mask in range(1, EVENT_MASK_NB):
                if trigger & event_mask:
                    triggers[variable, event_mask, indices[variable, event_mask]] = prop_idx
                    indices[variable, event_mask] += 1



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2026 - Yan Georget
###############################################################################
from nucs.examples.employee_scheduling.employee_scheduling_problem import EmployeeSchedulingProblem
from nucs.heuristics.heuristics import DOM_HEURISTIC_MAX_VALUE
from nucs.solvers.backtrack_solver import BacktrackSolver


class TestEmployeeScheduling:
    def test_nurses(self) -> None:
        assert list(EmployeeSchedulingProblem().nurses(0, 0)) == [0, 1, 2, 3, 4]
        assert list(EmployeeSchedulingProblem().nurses(0, 1)) == [5, 6, 7, 8, 9]
        assert list(EmployeeSchedulingProblem().nurses(1, 0)) == [15, 16, 17, 18, 19]
        assert list(EmployeeSchedulingProblem().nurses(1, 1)) == [20, 21, 22, 23, 24]

    def test_shifts(self) -> None:
        assert list(EmployeeSchedulingProblem().shifts(0, 0)) == [0, 5, 10]
        assert list(EmployeeSchedulingProblem().shifts(0, 1)) == [1, 6, 11]
        assert list(EmployeeSchedulingProblem().shifts(1, 0)) == [15, 20, 25]
        assert list(EmployeeSchedulingProblem().shifts(1, 1)) == [16, 21, 26]

    def test_employee_scheduling(self) -> None:
        problem = EmployeeSchedulingProblem()
        solver = BacktrackSolver(
            problem, decision_variables=problem.requested_shifts, dom_heuristic=DOM_HEURISTIC_MAX_VALUE
        )
        solution = solver.maximize(problem.satisfied_request_nb)
        assert solution is not None
        assert solution[problem.satisfied_request_nb] == 13



from nucs.problems.problem import Problem
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.propagators.propagators import ALG_ALLDIFFERENT

n = 8  # the number of queens
problem = Problem([(0, n - 1)] * n)
variables = range(0, n)
problem.add_propagator(ALG_ALLDIFFERENT, variables)
problem.add_propagator(ALG_ALLDIFFERENT, variables, range(n))
problem.add_propagator(ALG_ALLDIFFERENT, variables, range(0, -n, -1))
print(BacktrackSolver(problem).find_one()[:n])


###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2026 - Yan Georget
###############################################################################
import sys

from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import MAX, MIN
from nucs.heuristics.max_regret_var_heuristic import regret


@njit(cache=True)
def tsp_var_heuristic(decision_variables: NDArray, domains_stk: NDArray, stks_top: NDArray, params: NDArray) -> int:
    """
    :param decision_variables: the decision variables
    :param domains_stk: the stack of domains
    :param stks_top: the index of the top of the stacks as a Numpy array
    :param params: a two-dimensional (first dimension correspond to variables, second to values) cost array
    :return: the variable
    """
    best_score = -sys.maxsize
    best_variable = -1
    top = stks_top[0]
    for variable in decision_variables:
        domain = domains_stk[top, variable]
        if 0 < domain[MAX] - domain[MIN]:
            score = compute_score(domain, variable, params)
            if best_score < score:
                best_variable = variable
                best_score = score
    return best_variable


@njit(cache=True)
def compute_score(domain: NDArray, variable: int, params: NDArray) -> int:
    """
    Minimize [min(12, size(X)), -regret(X)] for lexicographic order.
    """
    size = min(12, domain[MAX] - domain[MIN] + 1)
    return -size * 1024 + regret(domain, params[variable])



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2026 - Yan Georget
###############################################################################
import argparse

from nucs.examples.default_argument_parser import DefaultArgumentParser
from nucs.problems.latin_square_problem import LatinSquareProblem, LatinSquareRCProblem
from nucs.solvers.backtrack_solver import BacktrackSolver

# Run with the following command (the second run is much faster because the code has been compiled):
# NUMBA_CACHE_DIR=.numba/cache python -m nucs.examples.latin_square -n 10
if __name__ == "__main__":
    parser = DefaultArgumentParser()
    parser.add_argument("-n", type=int, default=10)
    parser.add_argument("--rc", type=bool, action=argparse.BooleanOptionalAction, default=False)
    args = parser.parse_args()
    problem = LatinSquareRCProblem(args.n) if args.rc else LatinSquareProblem(range(args.n))
    solver = BacktrackSolver(problem, log_level=args.log_level, stks_max_height=args.cp_max_height)
    if args.find_all:
        solver.solve_all()
        if args.display_stats:
            solver.print_statistics()
    else:
        solution = solver.find_one()
        if args.display_stats:
            solver.print_statistics()
        if args.display_solutions:
            problem.print_solution(solution)



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2026 - Yan Georget
###############################################################################
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import EVENT_MASK_MIN_MAX, MAX, MIN, PROP_CONSISTENCY, PROP_INCONSISTENCY


def get_complexity_sum_eq(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return n


@njit(cache=True)
def get_triggers_sum_eq(n: int, variable: int, parameters: NDArray) -> int:
    """
    This propagator is triggered whenever there is a change in the domain of a variable.
    :param parameters: the parameters, unused here
    :return: an array of triggers
    """
    return EVENT_MASK_MIN_MAX


@njit(cache=True)
def compute_domains_sum_eq(domains: NDArray, parameters: NDArray) -> int:
    """
    Implements Sigma_i x_i = x_{n-1}.
    :param domains: the domains of the variables, x is an alias for domains
    :param parameters: the parameters of the propagator, unused here
    :return: the status of the propagation (consistency, inconsistency or entailment) as an int
    """
    n = len(domains) - 1
    domain_sum_min = -domains[-1, MIN] + domains[:n, MAX].sum()
    domain_sum_max = -domains[-1, MAX] + domains[:n, MIN].sum()
    for i in range(n):
        new_min = domains[i, MAX] - domain_sum_min
        new_max = domains[i, MIN] - domain_sum_max
        if new_min > domains[i, MIN]:
            domains[i, MIN] = new_min
        if new_max < domains[i, MAX]:
            domains[i, MAX] = new_max
        if domains[i, MIN] > domains[i, MAX]:
            return PROP_INCONSISTENCY
    new_min = domains[-1, MAX] + domain_sum_max
    new_max = domains[-1, MIN] + domain_sum_min
    if new_min > domains[-1, MIN]:
        domains[-1, MIN] = new_min
    if new_max < domains[-1, MAX]:
        domains[-1, MAX] = new_max
    if domains[-1, MIN] > domains[-1, MAX]:
        return PROP_INCONSISTENCY
    return PROP_CONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2026 - Yan Georget
###############################################################################
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import (
    EVENT_MASK_MIN,
    MAX,
    MIN,
    PROP_CONSISTENCY,
    PROP_ENTAILMENT,
    PROP_INCONSISTENCY,
)


def get_complexity_sum_leq_c(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return n


@njit(cache=True)
def get_triggers_sum_leq_c(n: int, variable: int, parameters: NDArray) -> int:
    """
    Returns the triggers for this propagator.
    :param parameters: the parameters
    :return: an array of triggers
    """
    return EVENT_MASK_MIN


@njit(cache=True)
def compute_domains_sum_leq_c(domains: NDArray, parameters: NDArray) -> int:
    """
    Implements Sigma_i x_i <= c.
    :param domains: the domains of the variables, x is an alias for domains
    :param parameters: the parameters of the propagator, c is the first parameter
    :return: the status of the propagation (consistency, inconsistency or entailment) as an int
    """
    n = len(domains)
    domain_sum_min = -parameters[0] + domains[:, MAX].sum()
    if domain_sum_min <= 0:
        return PROP_ENTAILMENT
    domain_sum_max = -parameters[0] + domains[:, MIN].sum()
    for i in range(n):
        new_max = domains[i, MIN] - domain_sum_max
        if new_max < domains[i, MAX]:
            domains[i, MAX] = new_max
        if domains[i, MIN] > domains[i, MAX]:
            return PROP_INCONSISTENCY
    return PROP_CONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2026 - Yan Georget
###############################################################################
import argparse

from nucs.constants import LOG_LEVEL_INFO, LOG_LEVELS, OPTIM_MODES, OPTIM_RESET
from nucs.solvers.consistency_algorithms import CONSISTENCY_ALG_BC


class DefaultArgumentParser(argparse.ArgumentParser):
    def __init__(self) -> None:
        super().__init__()
        self.add_argument(
            "--consistency",
            help="set the consistency algorithm (0 is for BC, 1 for BC+shaving)",
            type=int,
            default=CONSISTENCY_ALG_BC,
        )
        self.add_argument(
            "--cp-max-height",
            help="set the maximal height of the choice points stack (default is 512)",
            type=int,
            default=512,
        )
        self.add_argument(
            "--display-solutions",
            help="display the solution(s)",
            type=bool,
            action=argparse.BooleanOptionalAction,
            default=True,
        )
        self.add_argument(
            "--display-stats",
            help="display the statistics",
            type=bool,
            action=argparse.BooleanOptionalAction,
            default=True,
        )
        self.add_argument(
            "--find-all",
            help="find all solutions",
            type=bool,
            action=argparse.BooleanOptionalAction,
            default=False,
        )
        self.add_argument(
            "--ff",
            type=bool,
            action=argparse.BooleanOptionalAction,
            default=False,
        )
        self.add_argument(
            "--log-level",
            help="set the log level",
            choices=LOG_LEVELS,
            default=LOG_LEVEL_INFO,
        )
        self.add_argument(
            "--optimization-mode",
            help="set the optimization mode",
            choices=OPTIM_MODES,
            default=OPTIM_RESET,
        )
        self.add_argument(
            "--processors",
            help="set the number of processors",
            type=int,
            default=1,
        )
        self.add_argument(
            "--symmetry-breaking",
            help="add symmetry breaking constraints",
            action=argparse.BooleanOptionalAction,
            default=True,
        )



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2026 - Yan Georget
###############################################################################
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import (
    DOM_UPDATE_EVENTS,
    DOM_UPDATE_VARIABLE,
    EVENT_MASK_MAX,
    EVENT_MASK_MAX_GROUND,
    EVENT_MASK_MIN_GROUND,
    MAX,
    MIN,
)
from nucs.solvers.choice_points import cp_put


@njit(cache=True)
def max_value_dom_heuristic(
    domains_stk: NDArray,
    not_entailed_propagators_stk: NDArray,
    dom_update_stk: NDArray,
    stks_top: NDArray,
    variable: int,
    params: NDArray,
) -> int:
    """
    Chooses the max value of the domain.
    :param domains_stk: the stack of domains
    :param not_entailed_propagators_stk: the stack of not entailed propagators
    :param dom_update_stk: the stack of domain updates
    :param stks_top: the index of the top of the stacks as a Numpy array
    :param variable: the variable
    :param params: a two-dimensional parameters array, unused here
    :return: the events
    """
    top = stks_top[0]
    value = domains_stk[top, variable, MAX]
    cp_put(domains_stk, not_entailed_propagators_stk, stks_top)
    domains_stk[top + 1, variable, MIN] = value
    domains_stk[top, variable, MAX] = value - 1
    dom_update_stk[top, DOM_UPDATE_VARIABLE] = variable
    dom_update_stk[top, DOM_UPDATE_EVENTS] = (
        EVENT_MASK_MAX_GROUND if domains_stk[top, variable, MIN] == domains_stk[top, variable, MAX] else EVENT_MASK_MAX
    )
    return EVENT_MASK_MIN_GROUND
